import builtins
import types
import typing
from functools import partial
from typing import ClassVar, TypeVar, cast

import django
from django.core.exceptions import FieldDoesNotExist, FieldError
from django.core.serializers.python import Serializer as _PythonSerializer
from django.core.serializers.xml_serializer import Serializer as _XmlSerializer
from django.db import models
from django.db.models.base import DEFERRED, ModelBase  # type: ignore
from django.db.models.fields import Field
from django.db.models.options import Options, make_immutable_fields_list
from django.utils.encoding import smart_str
from typing_extensions import Self

if typing.TYPE_CHECKING:
    from django.db.models import Model, QuerySet
else:
    from django_stubs_ext import QuerySetAny as QuerySet

    reveal_type = print


T = TypeVar("T", bound="TypedModel", covariant=True)
TypedModelT = TypeVar("TypedModelT", bound="TypedModel")


class TypedModelManager(models.Manager[T]):
    def get_queryset(self) -> QuerySet[T]:
        qs = super().get_queryset()
        return self._filter_by_type(qs)

    def _filter_by_type(self, qs: QuerySet[T]) -> QuerySet[T]:
        if hasattr(self.model, "_typedmodels_type"):
            if self.model._typedmodels_subtypes and len(self.model._typedmodels_subtypes) > 1:
                qs = qs.filter(type__in=self.model._typedmodels_subtypes)
            else:
                qs = qs.filter(type=self.model._typedmodels_type)
        return qs


class TypedModelMetaclass(ModelBase):
    """
    This metaclass enables a model for auto-downcasting using a ``type`` attribute.
    """

    def __new__(meta, classname, bases, classdict) -> type["TypedModel"]:
        try:
            TypedModel  # noqa: B018
        except NameError:
            # don't do anything for TypedModel class itself
            #
            # ...except updating Meta class to instantiate fields_from_subclasses attribute
            typed_model = cast(
                type["TypedModel"],
                super().__new__(meta, classname, bases, classdict),
            )
            # We have to set this attribute after _meta has been created, otherwise an
            # exception would be thrown by Options class constructor.
            typed_model._meta.fields_from_subclasses = {}
            return typed_model

        # look for a non-proxy base class that is a subclass of TypedModel
        mro: list[type] = list(bases)
        base_class: type[TypedModel] | None
        while mro:
            base_class = mro.pop(-1)
            if issubclass(base_class, TypedModel) and base_class is not TypedModel:
                if base_class._meta.proxy or base_class._meta.abstract:
                    # continue up the mro looking for non-proxy base classes
                    mro.extend(base_class.__bases__)
                else:
                    break
        else:
            base_class = None

        if base_class:
            # Enforce that subclasses are proxy models.
            # Update an existing metaclass, or define an empty one
            # then set proxy=True
            class Meta:
                proxy: bool
                app_label: str

            Meta = classdict.get("Meta", Meta)  # type: ignore
            if getattr(Meta, "proxy", False):
                # If user has specified proxy=True explicitly, we assume that they want it to be treated like ordinary
                # proxy class, without TypedModel logic.
                return cast(
                    type["TypedModel"],
                    super().__new__(meta, classname, bases, classdict),
                )
            Meta.proxy = True

            # Proxy models shouldn't define their own indexes - they share the base model's table and indexes.
            # If we don't clear this, Django will complain that the index fields aren't local to the proxy model.
            # See issue #58
            if hasattr(Meta, "indexes"):
                # Set to empty list to override any inherited indexes
                # (delattr won't work for inherited attributes)
                Meta.indexes = []

            declared_fields = dict(
                (name, element)
                for name, element in list(classdict.items())
                if isinstance(element, Field)
            )

            for field_name, field in list(declared_fields.items()):
                # We need fields defined on subclasses to either:
                #  * be a ManyToManyField
                #  * have a default
                #  * be nullable
                if not (field.many_to_many or field.null or field.has_default()):
                    raise FieldError(
                        "All fields defined on typedmodels subclasses must be nullable, "
                        "or have a default set. "
                        f"Add null=True to the {classname}.{field_name} field definition."
                    )

                if isinstance(field, models.fields.related.RelatedField):
                    # Monkey patching field instance to make do_related_class use created class instead of base_class.
                    # Actually that class doesn't exist yet, so we just monkey patch base_class for a while,
                    # changing _meta.model_name, so accessor names are generated properly.
                    # We'll do more stuff when the class is created.
                    old_do_related_class = field.do_related_class

                    def do_related_class(self, other, cls):
                        base_class_name = base_class.__name__
                        cls._meta.model_name = classname.lower()
                        old_do_related_class(other, cls)  # noqa: B023
                        cls._meta.model_name = base_class_name.lower()

                    field.do_related_class = types.MethodType(do_related_class, field)  # type: ignore
                if isinstance(field, models.fields.related.RelatedField):
                    remote_field = field.remote_field
                    if isinstance(remote_field.model, TypedModel) and remote_field.model.base_class:
                        remote_field.limit_choices_to["type__in"] = (
                            remote_field.model._typedmodels_subtypes
                        )

                # Check if a field with this name has already been added to class
                try:
                    duplicate_field = base_class._meta.get_field(field_name)
                except FieldDoesNotExist:
                    field.contribute_to_class(base_class, field_name)
                else:
                    # Check if the field being added is _exactly_ the same as the field
                    # that already exists.
                    assert isinstance(duplicate_field, Field)
                    if duplicate_field.deconstruct()[1:] != field.deconstruct()[1:]:
                        raise ValueError(
                            "Can't add field '%s' from '%s' to '%s', field already exists.",
                            field_name,
                            classname,
                            base_class.__name__,
                        )

                classdict.pop(field_name)
            base_class._meta.fields_from_subclasses.update(declared_fields)

            # set app_label to the same as the base class, unless explicitly defined otherwise
            if not hasattr(Meta, "app_label"):
                if hasattr(getattr(base_class, "_meta", None), "app_label"):
                    Meta.app_label = base_class._meta.app_label

            classdict.update(
                {
                    "Meta": Meta,
                }
            )

        classdict["base_class"] = base_class

        cls = cast(
            type[TypedModel],
            super().__new__(meta, classname, bases, classdict),
        )

        cls._meta.fields_from_subclasses = {}

        if base_class:
            opts = cls._meta

            model_name = opts.model_name
            typ = f"{opts.app_label}.{model_name}"
            cls._typedmodels_type = typ
            cls._typedmodels_subtypes = [typ]
            if typ in base_class._typedmodels_registry:
                raise ValueError(
                    f"Can't register type {typ!r} to {classname!r} (already registered to {base_class._typedmodels_registry[typ].__name__!r})"
                )
            base_class._typedmodels_registry[typ] = cls

            type_name = getattr(cls._meta, "verbose_name", cls.__name__)
            type_field = cast(models.CharField, base_class._meta.get_field("type"))
            choices = (*(type_field.choices or ()), (typ, type_name))
            type_field.choices = sorted(choices)

            cls._meta.declared_fields = declared_fields

            # look for any other proxy superclasses, they'll need to know
            # about this subclass
            for superclass in cls.mro():
                if (
                    issubclass(superclass, base_class)
                    and superclass not in (cls, base_class)
                    and hasattr(superclass, "_typedmodels_type")
                ):
                    if superclass._typedmodels_subtypes is not None:
                        superclass._typedmodels_subtypes.append(typ)

            meta._patch_fields_cache(cls, base_class)
        elif not cls._meta.abstract:
            # this is the base class
            cls._typedmodels_registry = {}

            # Since fields may be added by subclasses, save original fields.
            cls._meta._typedmodels_original_fields = {f.name for f in cls._meta.fields}
            cls._meta._typedmodels_original_many_to_many = {f.name for f in cls._meta.many_to_many}

            # add a get_type_classes classmethod to allow fetching of all the subclasses (useful for admin)

            def _get_type_classes(subcls) -> list[type["TypedModel"]]:
                """
                Returns a list of the classes which are proxy subtypes of this concrete typed model.
                """
                if subcls is cls:
                    return list(cls._typedmodels_registry.values())
                else:
                    return [cls._typedmodels_registry[k] for k in subcls._typedmodels_subtypes]

            cls._get_type_classes = classmethod(_get_type_classes)  # type: ignore

            def _get_types(subcls) -> list[str]:
                """
                Returns a list of the possible string values (for the `type` attribute) for classes
                which are proxy subtypes of this concrete typed model.
                """
                if subcls is cls:
                    return list(cls._typedmodels_registry.keys())
                else:
                    return subcls._typedmodels_subtypes[:]

            cls.get_types = classmethod(_get_types)  # type: ignore

        return cls

    @staticmethod
    def _model_has_field(cls, base_class: type[TypedModelT], field_name: str):
        if field_name in base_class._meta._typedmodels_original_many_to_many:
            return True
        if field_name in base_class._meta._typedmodels_original_fields:
            return True
        if any(f.name == field_name for f in base_class._meta.private_fields):
            return True
        for ancestor in cls.mro():
            if issubclass(ancestor, base_class) and ancestor != base_class:
                if field_name in ancestor._meta.declared_fields.keys():
                    return True

        if field_name in cls._meta.fields_map:
            # Crazy case where a reverse M2M from another typedmodels proxy points to this proxy
            # (this is an m2m reverse field)
            return True
        return False

    @staticmethod
    def _patch_fields_cache(cls, base_class: type[TypedModelT]):
        orig_get_fields = cls._meta._get_fields

        if django.VERSION >= (5, 0):

            def _get_fields(
                self,
                forward=True,
                reverse=True,
                include_parents=True,
                include_hidden=False,
                topmost_call=True,
            ):
                cache_key = (
                    forward,
                    reverse,
                    include_parents,
                    include_hidden,
                    topmost_call,
                )

                was_cached = cache_key in self._get_fields_cache
                fields = orig_get_fields(
                    forward=forward,
                    reverse=reverse,
                    include_parents=include_parents,
                    include_hidden=include_hidden,
                    topmost_call=topmost_call,
                )
                # If it was cached already, it's because we've already filtered this, skip it
                if not was_cached:
                    fields = [
                        f
                        for f in fields
                        if TypedModelMetaclass._model_has_field(cls, base_class, f.name)
                    ]
                    fields = make_immutable_fields_list("get_fields()", fields)
                    self._get_fields_cache[cache_key] = fields
                return fields

        else:

            def _get_fields(
                self,
                forward=True,
                reverse=True,
                include_parents=True,
                include_hidden=False,
                seen_models=None,
            ):
                cache_key = (
                    forward,
                    reverse,
                    include_parents,
                    include_hidden,
                    seen_models is None,
                )

                was_cached = cache_key in self._get_fields_cache
                fields = orig_get_fields(
                    forward=forward,
                    reverse=reverse,
                    include_parents=include_parents,
                    include_hidden=include_hidden,
                    seen_models=seen_models,
                )
                # If it was cached already, it's because we've already filtered this, skip it
                if not was_cached:
                    fields = [
                        f
                        for f in fields
                        if TypedModelMetaclass._model_has_field(cls, base_class, f.name)
                    ]
                    fields = make_immutable_fields_list("get_fields()", fields)
                    self._get_fields_cache[cache_key] = fields
                return fields

        cls._meta._get_fields = partial(_get_fields, cls._meta)

        # If fields are already cached, expire the cache.
        cls._meta._expire_cache()


class TypedModelOptions(Options):
    _typedmodels_original_fields: set[str]
    _typedmodels_original_many_to_many: set[str]
    fields_from_subclasses: dict[str, Field]
    declared_fields: dict[str, Field]


class TypedModel(models.Model, metaclass=TypedModelMetaclass):
    '''
    This class contains the functionality required to auto-downcast a model based
    on its ``type`` attribute.

    To use, simply subclass TypedModel for your base type, and then subclass
    that for your concrete types.

    Example usage::

        from django.db import models
        from typedmodels.models import TypedModel

        class Animal(TypedModel):
            """
            Abstract model
            """
            name = models.CharField(max_length=255)

            def say_something(self):
                raise NotImplemented

            def __repr__(self):
                return u'<%s: %s>' % (self.__class__.__name__, self.name)

        class Canine(Animal):
            def say_something(self):
                return "woof"

        class Feline(Animal):
            def say_something(self):
                return "meoww"
    '''

    _typedmodels_type: ClassVar[str]
    _typedmodels_subtypes: ClassVar[list[str]]
    _typedmodels_registry: ClassVar[dict[str, type["TypedModel"]]]
    _meta: ClassVar[TypedModelOptions]

    objects: ClassVar[TypedModelManager[Self]] = TypedModelManager()

    type = models.CharField(choices=(), max_length=255, null=False, blank=False, db_index=True)

    # Class variable indicating if model should be automatically recasted after initialization
    _auto_recast = True

    class Meta:
        abstract = True

    @classmethod
    def from_db(cls, db, field_names, values):
        # Called when django instantiates a model class from a queryset.
        _typedmodels_do_recast = True
        if "type" not in field_names:
            # LIMITATION:
            # `type` was deferred in the queryset.
            # So we can't cast this model without generating another query.
            # That'd be *really* bad for performance.
            # Most likely, this would have happened in `obj.fieldname`, where `fieldname` is deferred.
            # This will populate a queryset with .only('fieldname'),
            # leading to this situation where 'type' is deferred.
            # Django will then copy the fieldname from those model instances onto the original obj.
            # In this case we don't really need a typed subclass, so we choose to just not recast in
            # this situation.
            #
            # Unfortunately we can't tell the difference between this situation and
            # MyModel.objects.only('myfield'). If you do that, we will also not recast.
            #
            # If you want you can recast manually (call obj.recast() on each object in your
            # queryset) - but by far a better solution would be to not defer the `type` field
            # to start with.
            _typedmodels_do_recast = False

        if len(values) != len(cls._meta.concrete_fields):
            values_iter = iter(values)
            values = [
                next(values_iter) if f.attname in field_names else DEFERRED
                for f in cls._meta.concrete_fields
            ]
        new = cls(*values, _typedmodels_do_recast=_typedmodels_do_recast)
        new._state.adding = False
        new._state.db = db
        return new

    @classmethod
    def get_type_classes(cls) -> list["__builtins__.type[Self]"]:
        """
        Returns a list of the classes which are proxy subtypes of this concrete typed model.
        """
        return cls._get_type_classes()  # type: ignore

    @classmethod
    def get_types(cls) -> list[str]:
        """
        Returns a list of the possible string values (for the `type` attribute) for classes
        which are proxy subtypes of this concrete typed model.
        """
        return cls._get_types()  # type: ignore

    def __init__(self, *args, _typedmodels_do_recast=None, **kwargs):
        # Calling __init__ on base class because some functions (e.g. save()) need access to field values from base
        # class.

        # Move args to kwargs since base_class may have more fields defined with different ordering
        args = list(args)
        if len(args) > len(self._meta.fields):
            # Daft, but matches old exception sans the err msg.
            raise IndexError("Number of args exceeds number of fields")
        for field_value, field in zip(args, self._meta.fields, strict=False):
            kwargs[field.attname] = field_value
        args = []  # args were all converted to kwargs

        if self.base_class:
            before_class = self.__class__
            self.__class__ = self.base_class
        else:
            before_class = None
        super().__init__(*args, **kwargs)
        if before_class:
            self.__class__ = before_class

        if _typedmodels_do_recast is None:
            _typedmodels_do_recast = self._auto_recast
        if _typedmodels_do_recast:
            self.recast()

    def recast(self, typ: builtins.type["TypedModel"] | None = None) -> None:
        for base in reversed(self.__class__.mro()):
            if issubclass(base, TypedModel) and hasattr(base, "_typedmodels_registry"):
                break
        else:
            raise ValueError("No suitable base class found to recast!")

        if not self.type:
            if not hasattr(self, "_typedmodels_type"):
                # This is an instance of an untyped model
                if typ is None:
                    # recast() is probably being called by __init__() here.
                    # Ideally we'd raise an error here, but the django admin likes to call
                    # model() and doesn't expect an error.
                    # Instead, we raise an error when the object is saved.
                    return
            else:
                self.type = self._typedmodels_type

        if typ is None:
            typ_str = self.type
        else:
            if isinstance(typ, type) and issubclass(typ, base):
                model_name = typ._meta.model_name
                typ_str = f"{typ._meta.app_label}.{model_name}"
            else:
                typ_str = str(typ)

        try:
            correct_cls = base._typedmodels_registry[typ_str]
        except KeyError:
            raise ValueError(f"Invalid {base.__name__} identifier: {typ!r}") from None

        self.type = typ_str

        current_cls = self.__class__

        if current_cls != correct_cls:
            self.__class__ = correct_cls

    def save(self, *args, **kwargs) -> None:
        self.presave(*args, **kwargs)
        return super().save(*args, **kwargs)

    def presave(self, *args, **kwargs) -> None:
        """Perform checks before saving the model."""
        if not getattr(self, "_typedmodels_type", None):
            raise RuntimeError(f"Untyped {self.__class__.__name__} cannot be saved.")

    def _get_unique_checks(self, exclude=None, **kwargs):
        unique_checks, date_checks = super()._get_unique_checks(exclude=exclude, **kwargs)

        for i, (_model_class, field_names) in reversed(list(enumerate(unique_checks))):
            for fn in field_names:
                try:
                    self._meta.get_field(fn)
                except FieldDoesNotExist:
                    # Some field in this unique check isn't actually on this proxy model.
                    unique_checks.pop(i)
                    break
        return unique_checks, date_checks


# Monkey patching Python and XML serializers in Django to use model name from base class.
# This should be preferably done by changing __unicode__ method for ._meta attribute in each model,
# but it doesn’t work.
_python_serializer_get_dump_object = _PythonSerializer.get_dump_object


def _get_dump_object(self, obj: "Model") -> dict:
    if isinstance(obj, TypedModel):
        return {
            "pk": smart_str(obj._get_pk_val(), strings_only=True),
            "model": smart_str(getattr(obj, "base_class", obj)._meta),
            "fields": self._current,
        }
    else:
        return _python_serializer_get_dump_object(self, obj)


_PythonSerializer.get_dump_object = _get_dump_object  # type: ignore

_xml_serializer_start_object = _XmlSerializer.start_object


def _start_object(self, obj: "Model") -> None:
    if isinstance(obj, TypedModel):
        self.indent(1)
        obj_pk = obj._get_pk_val()
        modelname = smart_str(getattr(obj, "base_class", obj)._meta)
        if obj_pk is None:
            attrs = {
                "model": modelname,
            }
        else:
            attrs = {
                "pk": smart_str(obj._get_pk_val()),
                "model": modelname,
            }

        self.xml.startElement("object", attrs)
    else:
        return _xml_serializer_start_object(self, obj)


_XmlSerializer.start_object = _start_object  # type: ignore



# Generated by Django 2.2.14 on 2020-07-26 18:18

from django.db import migrations, models
import django.db.models.deletion
import django.db.models.manager


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='AbstractVegetable',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[('testapp.fruit', 'fruit'), ('testapp.vegetable', 'vegetable')], db_index=True, max_length=255)),
                ('name', models.CharField(max_length=255)),
                ('color', models.CharField(max_length=255)),
                ('yumness', models.FloatField()),
            ],
            options={
                'abstract': False,
            },
            managers=[
                ('mymanager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='Animal',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[('testapp.angrybigcat', 'angry big cat'), ('testapp.bigcat', 'big cat'), ('testapp.canine', 'canine'), ('testapp.feline', 'feline'), ('testapp.parrot', 'parrot')], db_index=True, max_length=255)),
                ('name', models.CharField(max_length=255)),
                ('mice_eaten', models.IntegerField(default=0)),
                ('known_words', models.IntegerField(null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UniqueIdentifier',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('name', models.CharField(max_length=255)),
                ('content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='Parent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[('testapp.child1', 'child1'), ('testapp.child2', 'child2')], db_index=True, max_length=255)),
                ('a', models.CharField(max_length=1)),
                ('b', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='testapp.Parent')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Canine',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.animal',),
        ),
        migrations.CreateModel(
            name='Child1',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.parent',),
        ),
        migrations.CreateModel(
            name='Child2',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.parent',),
        ),
        migrations.CreateModel(
            name='Feline',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.animal',),
        ),
        migrations.CreateModel(
            name='Fruit',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.abstractvegetable',),
            managers=[
                ('mymanager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='Parrot',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.animal',),
        ),
        migrations.CreateModel(
            name='Vegetable',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.abstractvegetable',),
            managers=[
                ('mymanager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AddField(
            model_name='animal',
            name='canines_eaten',
            field=models.ManyToManyField(to='testapp.Canine'),
        ),
        migrations.CreateModel(
            name='BigCat',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.feline',),
        ),
        migrations.CreateModel(
            name='AngryBigCat',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.bigcat',),
        ),
    ]



# Migration for issue #58 - indexes on TypedModel base classes

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0002_developer_employee_manager'),
    ]

    operations = [
        migrations.CreateModel(
            name='BaseModelWithIndex',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[('testapp.submodela', 'Sub Model A'), ('testapp.submodelb', 'Sub Model B')], db_index=True, max_length=255)),
                ('name', models.CharField(max_length=100)),
                ('tag', models.CharField(max_length=100)),
                ('field_a', models.CharField(max_length=100, null=True)),
                ('field_b', models.IntegerField(null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddIndex(
            model_name='basemodelwithindex',
            index=models.Index(fields=['tag'], name='testapp_bas_tag_f26b85_idx'),
        ),
        migrations.CreateModel(
            name='SubModelA',
            fields=[
            ],
            options={
                'verbose_name': 'Sub Model A',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.basemodelwithindex',),
        ),
        migrations.CreateModel(
            name='SubModelB',
            fields=[
            ],
            options={
                'verbose_name': 'Sub Model B',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.basemodelwithindex',),
        ),
    ]



import pytest
from django.contrib.contenttypes.models import ContentType
from django.db import models

try:
    import yaml

    PYYAML_AVAILABLE = True
    del yaml
except ImportError:
    PYYAML_AVAILABLE = False

from django.core import serializers
from django.core.exceptions import FieldError

from testapp.models import (
    AbstractVegetable,
    AngryBigCat,
    Animal,
    BaseModelWithIndex,
    BigCat,
    Canine,
    Child2,
    Employee,
    Feline,
    Fruit,
    Parrot,
    SubModelA,
    SubModelB,
    UniqueIdentifier,
    Vegetable,
)

from .models import TypedModelManager


@pytest.fixture
def animals(db):
    kitteh = Feline.objects.create(name="kitteh")
    UniqueIdentifier.objects.create(
        name="kitteh",
        object_id=kitteh.pk,
        content_type=ContentType.objects.get_for_model(kitteh),
    )
    cheetah = Feline.objects.create(name="cheetah")
    UniqueIdentifier.objects.create(
        name="cheetah",
        object_id=cheetah.pk,
        content_type=ContentType.objects.get_for_model(cheetah),
    )
    fido = Canine.objects.create(name="fido")
    UniqueIdentifier.objects.create(
        name="fido",
        object_id=fido.pk,
        content_type=ContentType.objects.get_for_model(fido),
    )
    simba = BigCat.objects.create(name="simba")
    UniqueIdentifier.objects.create(
        name="simba",
        object_id=simba.pk,
        content_type=ContentType.objects.get_for_model(simba),
    )
    mufasa = AngryBigCat.objects.create(name="mufasa")
    UniqueIdentifier.objects.create(
        name="mufasa",
        object_id=mufasa.pk,
        content_type=ContentType.objects.get_for_model(mufasa),
    )
    kajtek = Parrot.objects.create(name="Kajtek")
    UniqueIdentifier.objects.create(
        name="kajtek",
        object_id=kajtek.pk,
        content_type=ContentType.objects.get_for_model(kajtek),
    )


def test_can_instantiate_base_model(db):
    # direct instantiation works fine without a type, as long as you don't save
    animal = Animal()
    assert not animal.type
    assert type(animal) is Animal


def test_cant_save_untyped_base_model(db):
    # direct instantiation shouldn't work
    with pytest.raises(RuntimeError):
        Animal.objects.create(name="uhoh")

    # ... unless a type is specified
    Animal.objects.create(name="dingo", type="testapp.canine")

    # ... unless that type is stupid
    with pytest.raises(ValueError):
        Animal.objects.create(name="dingo", type="macaroni.buffaloes")


def test_get_types():
    assert set(Animal.get_types()) == {
        "testapp.canine",
        "testapp.bigcat",
        "testapp.parrot",
        "testapp.angrybigcat",
        "testapp.feline",
    }
    assert set(Canine.get_types()) == {"testapp.canine"}
    assert set(Feline.get_types()) == {
        "testapp.bigcat",
        "testapp.angrybigcat",
        "testapp.feline",
    }


def test_get_type_classes():
    assert set(Animal.get_type_classes()) == {
        Canine,
        BigCat,
        Parrot,
        AngryBigCat,
        Feline,
    }
    assert set(Canine.get_type_classes()) == {Canine}
    assert set(Feline.get_type_classes()) == {BigCat, AngryBigCat, Feline}


def test_type_choices():
    type_choices = Animal._meta.get_field("type").choices
    assert type_choices == [
        ("testapp.angrybigcat", "angry big cat"),
        ("testapp.bigcat", "big cat"),
        ("testapp.canine", "canine"),
        ("testapp.feline", "feline"),
        ("testapp.parrot", "parrot"),
    ]
    assert {cls for cls, _ in type_choices} == set(Animal.get_types())


def test_base_model_queryset(animals):
    # all objects returned
    qs = Animal.objects.all().order_by("type")
    assert [obj.type for obj in qs] == [
        "testapp.angrybigcat",
        "testapp.bigcat",
        "testapp.canine",
        "testapp.feline",
        "testapp.feline",
        "testapp.parrot",
    ]
    assert [type(obj) for obj in qs] == [
        AngryBigCat,
        BigCat,
        Canine,
        Feline,
        Feline,
        Parrot,
    ]


def test_proxy_model_queryset(animals):
    qs = Canine.objects.all().order_by("type")
    assert qs.count() == 1
    assert len(qs) == 1
    assert [obj.type for obj in qs] == ["testapp.canine"]
    assert [type(obj) for obj in qs] == [Canine]

    qs = Feline.objects.all().order_by("type")
    assert qs.count() == 4
    assert len(qs) == 4
    assert [obj.type for obj in qs] == [
        "testapp.angrybigcat",
        "testapp.bigcat",
        "testapp.feline",
        "testapp.feline",
    ]
    assert [type(obj) for obj in qs] == [AngryBigCat, BigCat, Feline, Feline]


def test_doubly_proxied_model_queryset(animals):
    qs = BigCat.objects.all().order_by("type")
    assert qs.count() == 2
    assert len(qs) == 2
    assert [obj.type for obj in qs] == ["testapp.angrybigcat", "testapp.bigcat"]
    assert [type(obj) for obj in qs] == [AngryBigCat, BigCat]


def test_triply_proxied_model_queryset(animals):
    qs = AngryBigCat.objects.all().order_by("type")
    assert qs.count() == 1
    assert len(qs) == 1
    assert [obj.type for obj in qs] == ["testapp.angrybigcat"]
    assert [type(obj) for obj in qs] == [AngryBigCat]


def test_recast_auto(animals):
    cat = Feline.objects.get(name="kitteh")
    cat.type = "testapp.bigcat"
    cat.recast()
    assert cat.type == "testapp.bigcat"
    assert type(cat) == BigCat


def test_recast_to_subclass_string(animals):
    cat = Feline.objects.get(name="kitteh")
    cat.recast("testapp.bigcat")
    assert cat.type == "testapp.bigcat"
    assert type(cat) == BigCat


def test_recast_to_subclass_modelclass(animals):
    cat = Feline.objects.get(name="kitteh")
    cat.recast(BigCat)
    assert cat.type == "testapp.bigcat"
    assert type(cat) == BigCat


def test_recast_string(animals):
    cat = Feline.objects.get(name="kitteh")
    cat.recast("testapp.canine")
    assert cat.type == "testapp.canine"
    assert type(cat) == Canine


def test_recast_modelclass(animals):
    cat = Feline.objects.get(name="kitteh")
    cat.recast(Canine)
    assert cat.type == "testapp.canine"
    assert type(cat) == Canine


def test_recast_fail(animals):
    cat = Feline.objects.get(name="kitteh")
    with pytest.raises(ValueError):
        cat.recast(AbstractVegetable)
    with pytest.raises(ValueError):
        cat.recast("typedmodels.abstractvegetable")
    with pytest.raises(ValueError):
        cat.recast(Vegetable)
    with pytest.raises(ValueError):
        cat.recast("typedmodels.vegetable")


def test_fields_in_subclasses(animals):
    canine = Canine.objects.all()[0]
    angry = AngryBigCat.objects.all()[0]

    angry.mice_eaten = 5
    angry.save()
    assert AngryBigCat.objects.get(pk=angry.pk).mice_eaten == 5

    angry.canines_eaten.add(canine)
    assert list(angry.canines_eaten.all()) == [canine]

    # Feline class was created before Parrot and has mice_eaten field which is non-m2m, so it may break accessing
    # known_words field in Parrot instances (since Django 1.5).
    parrot = Parrot.objects.all()[0]
    parrot.known_words = 500
    parrot.save()
    assert Parrot.objects.get(pk=parrot.pk).known_words == 500


def test_fields_cache():
    mice_eaten = Feline._meta.get_field("mice_eaten")
    known_words = Parrot._meta.get_field("known_words")
    assert mice_eaten in AngryBigCat._meta.fields
    assert mice_eaten in Feline._meta.fields
    assert mice_eaten not in Parrot._meta.fields
    assert known_words in Parrot._meta.fields
    assert known_words not in AngryBigCat._meta.fields
    assert known_words not in Feline._meta.fields


def test_m2m_cache():
    canines_eaten = AngryBigCat._meta.get_field("canines_eaten")
    assert canines_eaten in AngryBigCat._meta.many_to_many
    assert canines_eaten not in Feline._meta.many_to_many
    assert canines_eaten not in Parrot._meta.many_to_many


def test_related_names(animals):
    """Ensure that accessor names for reverse relations are generated properly."""

    canine = Canine.objects.all()[0]
    assert hasattr(canine, "angrybigcat_set")


def test_queryset_defer(db):
    """
    Ensure that qs.defer() works correctly
    """
    Vegetable.objects.create(name="cauliflower", color="white", yumness=1)
    Vegetable.objects.create(name="spinach", color="green", yumness=5)
    Vegetable.objects.create(name="sweetcorn", color="yellow", yumness=10)
    Fruit.objects.create(name="Apple", color="red", yumness=7)

    qs = AbstractVegetable.objects.defer("yumness")

    objs = set(qs)
    for o in objs:
        assert isinstance(o, AbstractVegetable)
        assert set(o.get_deferred_fields()) == {"yumness"}
        # does a query, since this field was deferred
        assert isinstance(o.yumness, float)


def test_queryset_defer_type(db):
    Vegetable.objects.create(name="cauliflower", color="white", yumness=1)
    Fruit.objects.create(name="Apple", color="red", yumness=7)

    qs = AbstractVegetable.objects.only("id")
    assert len(qs) == 2
    assert type(qs[0]) is AbstractVegetable
    assert type(qs[1]) is AbstractVegetable

    qs = Vegetable.objects.only("id")
    assert len(qs) == 1
    assert type(qs[0]) is Vegetable


def test_queryset_defer_type_with_subclass_fields(db, animals):
    # <relatedQuerySet>.delete() tends to do this .only('id') thing while collating referenced models.
    # So it needs to work even if you think it's a weird thing to do
    # In this case we *don't* auto-cast to anything; all returned models are Animal instances
    # this avoids the following error:
    #     django.core.exceptions.FieldDoesNotExist: AngryBigCat has no field named 'known_words'
    qs = list(Animal.objects.only("id"))
    assert len(qs) == 6
    assert all(type(x) is Animal for x in qs)


@pytest.mark.parametrize(
    "fmt",
    [
        "xml",
        "json",
        pytest.param(
            "yaml",
            marks=[pytest.mark.skipif(not PYYAML_AVAILABLE, reason="PyYAML is not available")],
        ),
    ],
)
def test_serialization(fmt, animals):
    """Helper function used to check serialization and deserialization for concrete format."""
    animals = Animal.objects.order_by("pk")
    serialized_animals = serializers.serialize(fmt, animals)
    deserialized_animals = [
        wrapper.object for wrapper in serializers.deserialize(fmt, serialized_animals)
    ]
    assert set(deserialized_animals) == set(animals)


def test_generic_relation(animals):
    for animal in Animal.objects.all():
        assert hasattr(animal, "unique_identifiers")
        assert animal.unique_identifiers.all()

    Feline._meta.get_field("unique_identifiers")
    for feline in Feline.objects.all():
        assert hasattr(feline, "unique_identifiers")
        assert feline.unique_identifiers.all()

    for uid in UniqueIdentifier.objects.all():
        cls = uid.referent.__class__
        animal = cls.objects.filter(unique_identifiers=uid)
        assert isinstance(animal.first(), Animal)

    for uid in UniqueIdentifier.objects.all():
        cls = uid.referent.__class__
        animal = cls.objects.filter(unique_identifiers__name=uid.name)
        assert isinstance(animal.first(), Animal)


def test_manager_classes():
    assert isinstance(Animal.objects, TypedModelManager)
    assert isinstance(Feline.objects, TypedModelManager)
    assert isinstance(BigCat.objects, TypedModelManager)

    # This one has a custom manager defined, but that shouldn't prevent objects from working
    assert isinstance(AbstractVegetable.mymanager, models.Manager)
    assert isinstance(AbstractVegetable.objects, TypedModelManager)

    # subclasses work the same way
    assert isinstance(Vegetable.mymanager, models.Manager)
    assert isinstance(Vegetable.objects, TypedModelManager)


def test_uniqueness_check_on_child(db):
    child2 = Child2.objects.create(a="a")

    # Regression test for https://github.com/craigds/django-typed-models/issues/42
    # FieldDoesNotExist: Child2 has no field named 'b'
    child2.validate_unique()


def test_uniqueness_check_include_meta_constraints(db):
    """Django 4.1 introduces a new required kwarg
    django/forms/models.py in validate_unique at line 809
    for form in valid_forms:
    exclude = form._get_validation_exclusions()
    unique_checks, date_checks = form.instance._get_unique_checks(
        exclude=exclude,
        include_meta_constraints=True,
    )
    """
    child2 = Child2.objects.create(a="a")
    child2._get_unique_checks()


def test_non_nullable_subclass_field_error():
    with pytest.raises(FieldError):

        class Bug(Animal):
            # should have null=True
            num_legs = models.PositiveIntegerField()


def test_explicit_recast_with_class_on_untyped_instance():
    animal = Animal()
    assert not animal.type
    animal.recast(Feline)
    assert animal.type == "testapp.feline"
    assert type(animal) is Feline


def test_explicit_recast_with_string_on_untyped_instance():
    animal = Animal()
    assert not animal.type
    animal.recast("testapp.feline")
    assert animal.type == "testapp.feline"
    assert type(animal) is Feline


def test_same_field_name_in_two_subclasses():
    with pytest.raises(ValueError):

        class Tester1(Employee):
            name = models.CharField(max_length=255, blank=True, null=True)

    with pytest.raises(ValueError):

        class Tester2(Employee):
            name = models.CharField(max_length=254, null=True)

    with pytest.raises(ValueError):

        class Tester3(Employee):
            name = models.IntegerField(null=True)


def test_base_model_with_indexes(db):
    """Test that indexes on base TypedModel work correctly (issue #58).

    The issue: When a TypedModel base class has indexes in Meta, and proxy
    subclasses inherit from that Meta, they would inherit the indexes attribute.
    Django's system checks would then fail because the indexed fields aren't
    local to the proxy models.

    The fix: TypedModelMetaclass removes the 'indexes' attribute from the Meta
    class before Django's ModelBase processes proxy models, preventing the error.
    """
    # Verify that SubModelA and SubModelB are proxy models
    assert SubModelA._meta.proxy is True
    assert SubModelB._meta.proxy is True

    # The base model should have the index in its _meta
    assert (
        len(BaseModelWithIndex._meta.indexes) == 1
    ), "Base model should have its index defined in _meta"

    # Proxy models should NOT have indexes in their _meta (they share the base table)
    assert (
        len(SubModelA._meta.indexes) == 0
    ), "Proxy models shouldn't have indexes in _meta (they use the base model's table)"
    assert (
        len(SubModelB._meta.indexes) == 0
    ), "Proxy models shouldn't have indexes in _meta (they use the base model's table)"

    # Verify the models work at runtime
    obj_a = SubModelA.objects.create(name="test_a", tag="tag1", field_a="a_value")
    obj_b = SubModelB.objects.create(name="test_b", tag="tag2", field_b=42)

    assert SubModelA.objects.get(pk=obj_a.pk).field_a == "a_value"
    assert SubModelB.objects.get(pk=obj_b.pk).field_b == 42
    assert SubModelA.objects.count() == 1
    assert SubModelB.objects.count() == 1
    assert BaseModelWithIndex.objects.count() == 2



from typing import TYPE_CHECKING, Generic

from django.contrib.admin import ModelAdmin

from .models import TypedModel, TypedModelT

if TYPE_CHECKING:
    from django.forms.forms import BaseForm
    from django.http import HttpRequest


class TypedModelAdmin(ModelAdmin, Generic[TypedModelT]):
    model: "type[TypedModelT]"

    def get_fields(
        self,
        request: "HttpRequest",
        obj: "TypedModelT | None" = None,
    ) -> list[str | list[str] | tuple[str, ...]]:
        fields = list(super().get_fields(request, obj))
        # we remove the type field from the admin of subclasses.
        if TypedModel not in self.model.__bases__:
            fields.remove(self.model._meta.get_field("type").name)
        return fields

    def save_model(
        self,
        request: "HttpRequest",
        obj: "TypedModelT",
        form: "BaseForm",
        change,
    ) -> None:
        if getattr(obj, "_typedmodels_type", None) is None:
            # new instances don't have the type attribute
            obj._typedmodels_type = form.cleaned_data["type"]  # type: ignore[misc]
        obj.save()



INSTALLED_APPS = (
    "typedmodels",
    "django.contrib.contenttypes",
    "testapp",
)
MIDDLEWARE_CLASSES = ()
DATABASES = {"default": {"NAME": ":memory:", "ENGINE": "django.db.backends.sqlite3"}}
SECRET_KEY = "abc123"
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
# avoid RemovedInDjango50Warning when running tests:
USE_TZ = True



# Generated by Django 3.2.6 on 2021-08-19 17:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Employee',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[('testapp.developer', 'developer'), ('testapp.manager', 'manager')], db_index=True, max_length=255)),
                ('name', models.CharField(max_length=255, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Developer',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.employee',),
        ),
        migrations.CreateModel(
            name='Manager',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.employee',),
        ),
    ]



from typing import ClassVar

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import CharField, ForeignKey, PositiveIntegerField

from typedmodels.models import TypedModel, TypedModelManager


class UniqueIdentifier(models.Model):
    referent = GenericForeignKey()
    content_type = ForeignKey(ContentType, null=True, blank=True, on_delete=models.CASCADE)
    object_id = PositiveIntegerField(null=True, blank=True)
    created = models.DateTimeField(db_index=True, auto_now_add=True)
    name = CharField(max_length=255)


class UniqueIdentifierMixin(models.Model):
    unique_identifiers = GenericRelation(UniqueIdentifier, related_query_name="referents")

    class Meta:
        abstract = True


class Animal(TypedModel, UniqueIdentifierMixin):
    """
    Abstract model
    """

    name = models.CharField(max_length=255)

    def say_something(self):
        raise NotImplementedError

    # def __repr__(self):
    #     return u'<%s: %s>' % (self.__class__.__name__, self.name)

    def __str__(self):
        return str(self.name)


class Canine(Animal):
    def say_something(self):
        return "woof"


class Feline(Animal):
    def say_something(self):
        return "meoww"

    mice_eaten = models.IntegerField(default=0)


class BigCat(Feline):
    """
    This model tests doubly-proxied models.
    """

    def say_something(self):
        return "roar"


class AngryBigCat(BigCat):
    """
    This model tests triple-proxied models. Because we can
    """

    canines_eaten = models.ManyToManyField(Canine)

    def say_something(self):
        return "raawr"


class Parrot(Animal):
    known_words = models.IntegerField(null=True)

    def say_something(self):
        return "hello"


class AbstractVegetable(TypedModel):
    """
    This is an entirely different typed model.
    """

    name = models.CharField(max_length=255)
    color = models.CharField(max_length=255)
    yumness = models.FloatField(null=False)

    mymanager = models.Manager()


class Fruit(AbstractVegetable):
    pass


class Vegetable(AbstractVegetable):
    pass


class SurpriseAbstractModel(TypedModel):
    """
    This class *isn't* the typed base, it's a random abstract model.
    The presence of this model tests
    https://github.com/craigds/django-typed-models/issues/61
    """

    class Meta:
        abstract = True


class Parent(SurpriseAbstractModel):
    a = models.CharField(max_length=1)


class Child1(Parent):
    b = models.OneToOneField("self", null=True, on_delete=models.CASCADE)


class Child2(Parent):
    pass


class EmployeeManager(TypedModelManager):
    pass


class Employee(TypedModel):
    objects: ClassVar[EmployeeManager] = EmployeeManager()


class Developer(Employee):
    name = models.CharField(max_length=255, null=True)


class Manager(Employee):
    # Adds the _exact_ same field as Developer. Shouldn't error.
    name = models.CharField(max_length=255, null=True)


def typed_queryset() -> None:
    # This isn't actually called, but it's here for the mypy check to ensure that type hinting works correctly.
    queryset = Animal.objects.filter(pk=1)
    queryset.filter(name="lynx")  # works, because Animal has this field


def do_get_type_classes() -> None:
    for x in Animal.get_type_classes():
        print(x)


# Test model for issue #58 - indexes on TypedModel base classes
# The issue occurs when:
# 1. An abstract model defines indexes
# 2. A concrete TypedModel inherits from that abstract model
# 3. Proxy subclasses of the TypedModel inherit the Meta with indexes


class AbstractModelWithIndex(models.Model):
    """Abstract model with indexes."""

    name = models.CharField(max_length=100)
    tag = models.CharField(max_length=100)

    class Meta:
        abstract = True
        indexes = [models.Index(fields=["tag"])]


class BaseModelWithIndex(AbstractModelWithIndex, TypedModel):
    """Concrete TypedModel inheriting from abstract model."""

    pass


class SubModelA(BaseModelWithIndex):
    field_a = models.CharField(max_length=100, null=True)

    class Meta(BaseModelWithIndex.Meta):
        verbose_name = "Sub Model A"


class SubModelB(BaseModelWithIndex):
    field_b = models.IntegerField(null=True)

    class Meta(BaseModelWithIndex.Meta):
        verbose_name = "Sub Model B"
